- Add BF16 support for SM90 and SM100 - Refactor Python APIs - Other fixes and code refactoring
344 lines
17 KiB
Plaintext
344 lines
17 KiB
Plaintext
#pragma once
|
|
|
|
#pragma clang diagnostic push
|
|
#pragma clang diagnostic ignored "-Wunknown-attributes"
|
|
|
|
#include <cutlass/arch/barrier.h>
|
|
#include <cutlass/arch/reg_reconfig.h>
|
|
|
|
#include <cute/arch/cluster_sm90.hpp>
|
|
#include <cute/arch/copy_sm90_desc.hpp>
|
|
#include <cute/arch/copy_sm90_tma.hpp>
|
|
|
|
#include <deep_gemm/common/utils.cuh>
|
|
#include <deep_gemm/common/scheduler.cuh>
|
|
#include <deep_gemm/common/sm90_utils.cuh>
|
|
|
|
namespace deep_gemm {
|
|
|
|
using namespace deep_gemm::sm90;
|
|
|
|
template <uint32_t SHAPE_M, uint32_t SHAPE_N, uint32_t SHAPE_K,
|
|
uint32_t kNumGroups,
|
|
uint32_t BLOCK_M, uint32_t BLOCK_N, uint32_t BLOCK_K,
|
|
uint32_t kSwizzleDMode,
|
|
uint32_t kNumStages, uint32_t kNumLastStages,
|
|
uint32_t kNumTMAThreads, uint32_t kNumMathThreads,
|
|
uint32_t kNumTMAMulticast, bool kIsTMAMulticastOnA,
|
|
uint32_t kNumSMs, GemmType kGemmType>
|
|
__global__ __launch_bounds__(kNumTMAThreads + kNumMathThreads, 1) void
|
|
sm90_bf16_gemm_impl(int* grouped_layout,
|
|
uint32_t shape_m, uint32_t shape_n, uint32_t shape_k,
|
|
const __grid_constant__ cute::TmaDescriptor tensor_map_a,
|
|
const __grid_constant__ cute::TmaDescriptor tensor_map_b,
|
|
const __grid_constant__ cute::TmaDescriptor tensor_map_d) {
|
|
#if (defined(__CUDA_ARCH__) and (__CUDA_ARCH__ >= 900)) or defined(__CLION_IDE__)
|
|
// Types
|
|
using WGMMA = typename BF16MMASelector<BLOCK_N>::type;
|
|
using Barrier = cutlass::arch::ClusterTransactionBarrier;
|
|
DG_STATIC_ASSERT(BLOCK_M % WGMMA::M == 0, "Invalid block size");
|
|
|
|
// Overwrite shape constants if the compiler gives
|
|
shape_m = SHAPE_M != 0 ? SHAPE_M : shape_m;
|
|
shape_n = SHAPE_N != 0 ? SHAPE_N : shape_n;
|
|
shape_k = SHAPE_K != 0 ? SHAPE_K : shape_k;
|
|
|
|
// Shared memory
|
|
static constexpr uint32_t SMEM_D_SIZE = BLOCK_M * BLOCK_N * sizeof(__nv_bfloat16);
|
|
static constexpr uint32_t SMEM_A_SIZE_PER_STAGE = BLOCK_M * BLOCK_K * sizeof(__nv_bfloat16);
|
|
static constexpr uint32_t SMEM_B_SIZE_PER_STAGE = BLOCK_N * BLOCK_K * sizeof(__nv_bfloat16);
|
|
|
|
// Configs
|
|
constexpr uint32_t kFullKOfAllStages = kNumStages * BLOCK_K;
|
|
const uint32_t num_iterations = ceil_div(shape_k, kFullKOfAllStages);
|
|
const uint32_t warp_idx = __shfl_sync(0xffffffff, threadIdx.x / 32, 0);
|
|
const uint32_t lane_idx = get_lane_idx();
|
|
|
|
// Prefetch TMA descriptors at the very beginning
|
|
if (threadIdx.x == kNumMathThreads) {
|
|
cute::prefetch_tma_descriptor(&tensor_map_a);
|
|
cute::prefetch_tma_descriptor(&tensor_map_b);
|
|
cute::prefetch_tma_descriptor(&tensor_map_d);
|
|
}
|
|
__syncwarp();
|
|
|
|
// Align to 1024 bytes for swizzle-128B
|
|
extern __shared__ __align__(1024) uint8_t smem_buffer[];
|
|
DG_STATIC_ASSERT(SMEM_D_SIZE % 1024 == 0, "Shared memory of A/B must be aligned to 1024 bytes");
|
|
|
|
// Data on shared memory
|
|
auto smem_d = reinterpret_cast<__nv_bfloat16*>(smem_buffer);
|
|
__nv_bfloat16* smem_a[kNumStages];
|
|
__nv_bfloat16* smem_b[kNumStages];
|
|
|
|
// TMA Barrier for both divisible and non-divisible cases
|
|
Barrier* full_barriers[kNumStages];
|
|
Barrier* empty_barriers[kNumStages];
|
|
|
|
// Fill shared memory pointers
|
|
#pragma unroll
|
|
for (uint32_t i = 0; i < kNumStages; ++ i) {
|
|
smem_a[i] = reinterpret_cast<__nv_bfloat16*>(smem_buffer + SMEM_D_SIZE + i * SMEM_A_SIZE_PER_STAGE);
|
|
smem_b[i] = reinterpret_cast<__nv_bfloat16*>(smem_buffer + SMEM_D_SIZE + kNumStages * SMEM_A_SIZE_PER_STAGE + i * SMEM_B_SIZE_PER_STAGE);
|
|
}
|
|
|
|
// Fill barriers
|
|
auto barrier_start_ptr = reinterpret_cast<Barrier*>(smem_buffer + SMEM_D_SIZE + kNumStages * (SMEM_A_SIZE_PER_STAGE + SMEM_B_SIZE_PER_STAGE));
|
|
#pragma unroll
|
|
for (uint32_t i = 0; i < kNumStages; ++ i) {
|
|
full_barriers[i] = barrier_start_ptr + i;
|
|
empty_barriers[i] = barrier_start_ptr + kNumStages + i;
|
|
}
|
|
|
|
// Initialize barriers
|
|
if (threadIdx.x == kNumMathThreads) {
|
|
#pragma unroll
|
|
for (uint32_t i = 0; i < kNumStages; ++ i) {
|
|
full_barriers[i]->init(1);
|
|
empty_barriers[i]->init(kNumTMAMulticast * kNumMathThreads / 32);
|
|
}
|
|
|
|
// Make initialized barrier visible in async proxy
|
|
cutlass::arch::fence_view_async_shared();
|
|
cutlass::arch::fence_barrier_init();
|
|
}
|
|
|
|
// Synchronize all threads to make barrier visible in normal memory model
|
|
(kNumTMAMulticast > 1) ? cute::cluster_sync() : __syncthreads();
|
|
|
|
struct DivisibleK {};
|
|
struct NotDivisibleK {};
|
|
auto launch_k_iterations = [=](const auto& func) {
|
|
if constexpr (kNumLastStages == 0) {
|
|
for (uint32_t k_iter = 0; k_iter < num_iterations; ++ k_iter)
|
|
func(k_iter, DivisibleK{});
|
|
} else {
|
|
for (uint32_t k_iter = 0; k_iter < num_iterations - 1; ++ k_iter)
|
|
func(k_iter, DivisibleK{});
|
|
func(num_iterations - 1, NotDivisibleK{});
|
|
}
|
|
};
|
|
|
|
// Register reconfigurations
|
|
constexpr uint32_t kNumTMARegisters = 48;
|
|
constexpr uint32_t kNumMathRegisters = 224;
|
|
|
|
// Block scheduler
|
|
uint32_t m_block_idx, n_block_idx;
|
|
auto scheduler = Scheduler<kGemmType, BLOCK_M, BLOCK_N, kNumGroups, kNumTMAMulticast, kIsTMAMulticastOnA, kNumSMs>(shape_m, shape_n, grouped_layout);
|
|
|
|
if (threadIdx.x >= kNumMathThreads) {
|
|
// TMA warp-group for loading data
|
|
cutlass::arch::warpgroup_reg_dealloc<kNumTMARegisters>();
|
|
|
|
// NOTES: only one thread (or warp) will be used
|
|
if (threadIdx.x < kNumMathThreads + 32 and cute::elect_one_sync()) {
|
|
// Persistently schedule over blocks
|
|
while (scheduler.get_next_block(m_block_idx, n_block_idx)) {
|
|
launch_k_iterations([&](uint32_t k_iter, auto divisible_type) {
|
|
constexpr bool kHasDivisibleStages = cute::is_same_v<decltype(divisible_type), DivisibleK>;
|
|
constexpr uint32_t kNumInnerStages = kHasDivisibleStages ? kNumStages : kNumLastStages;
|
|
|
|
// Assign TMA multicast number into A and B
|
|
// NOTES: there may be additional odd rows/columns or cases where multicast is not possible.
|
|
const bool is_tma_multicast_valid = scheduler.is_tma_multicast_valid(m_block_idx);
|
|
const uint32_t num_tma_multicast_a = (kIsTMAMulticastOnA and is_tma_multicast_valid) ? kNumTMAMulticast : 1;
|
|
const uint32_t num_tma_multicast_b = (not kIsTMAMulticastOnA and is_tma_multicast_valid) ? kNumTMAMulticast : 1;
|
|
DG_STATIC_ASSERT(kNumTMAMulticast <= 2, "Scheduler does not support > 2 TMA multicast");
|
|
|
|
// NOTES: unrolling and `kNumInnerStages` are vital for performance, NVCC will try to eliminate all
|
|
// shared memory pointers, e.g. `full_barriers` registers, if all the access indices are constant
|
|
#pragma unroll
|
|
for (uint32_t s = 0; s < kNumInnerStages; ++ s) {
|
|
// Wait consumer release
|
|
empty_barriers[s]->wait((scheduler.current_iter * num_iterations + k_iter + 1) & 1);
|
|
|
|
constexpr bool kWithGroupOffsetA = kGemmType == GemmType::MGroupedMasked;
|
|
auto& full_barrier = *full_barriers[s];
|
|
uint32_t k_idx = k_iter * kFullKOfAllStages + s * BLOCK_K;
|
|
|
|
tma_copy(&tensor_map_a, reinterpret_cast<uint64_t*>(&full_barrier),
|
|
smem_a[s], k_idx, scheduler.get_global_idx<kWithGroupOffsetA>(shape_m, BLOCK_M, m_block_idx),
|
|
num_tma_multicast_a);
|
|
tma_copy(&tensor_map_b, reinterpret_cast<uint64_t*>(&full_barrier),
|
|
smem_b[s], k_idx, scheduler.get_global_idx<true>(shape_n, BLOCK_N, n_block_idx, m_block_idx),
|
|
num_tma_multicast_b);
|
|
full_barrier.arrive_and_expect_tx(SMEM_A_SIZE_PER_STAGE + SMEM_B_SIZE_PER_STAGE);
|
|
}
|
|
|
|
#pragma unroll
|
|
for (uint32_t s = kNumInnerStages; s < kNumStages; ++ s) {
|
|
empty_barriers[s]->wait((scheduler.current_iter * num_iterations + k_iter + 1) & 1);
|
|
full_barriers[s]->arrive();
|
|
}
|
|
});
|
|
}
|
|
|
|
// To safely deconstruct distributed shared barriers, we need another round of empty waits
|
|
if constexpr (kNumTMAMulticast > 1) {
|
|
#pragma unroll
|
|
for (uint32_t s = 0; s < kNumStages; ++ s)
|
|
empty_barriers[s]->wait((scheduler.current_iter * num_iterations + 1) & 1);
|
|
}
|
|
}
|
|
} else {
|
|
// Math warp-groups for WGMMA
|
|
cutlass::arch::warpgroup_reg_alloc<kNumMathRegisters>();
|
|
|
|
// NOTES: use `__shfl_sync` to encourage NVCC to use unified registers
|
|
const auto math_wg_idx = __shfl_sync(0xffffffff, threadIdx.x / 128, 0);
|
|
|
|
while (scheduler.get_next_block(m_block_idx, n_block_idx)) {
|
|
constexpr uint32_t WAVE_BLOCK_M = WGMMA::M * (BLOCK_M <= 64 ? 1 : 2);
|
|
DG_STATIC_ASSERT(BLOCK_M % WAVE_BLOCK_M == 0, "Invalid block sizes");
|
|
float accum[WGMMA::kNumAccum * (BLOCK_M / WAVE_BLOCK_M)] = {0};
|
|
|
|
// Empty barrier arrival
|
|
auto empty_barrier_arrive = [&](uint32_t s) {
|
|
if constexpr (kNumTMAMulticast == 1) {
|
|
lane_idx == 0 ? empty_barriers[s]->arrive() : void();
|
|
} else {
|
|
auto target_cta = scheduler.is_peer_cta_alive ? lane_idx : cute::block_rank_in_cluster();
|
|
lane_idx < kNumTMAMulticast ? empty_barriers[s]->arrive(target_cta) : void();
|
|
}
|
|
};
|
|
|
|
cutlass::arch::NamedBarrier(kNumMathThreads).sync();
|
|
|
|
// Launch MMAs
|
|
launch_k_iterations([&](uint32_t k_iter, auto divisible_type) {
|
|
constexpr bool kHasDivisibleStages = cute::is_same_v<decltype(divisible_type), DivisibleK>;
|
|
constexpr uint32_t kNumInnerStages = kHasDivisibleStages ? kNumStages : kNumLastStages;
|
|
|
|
// TODO: remove some useless computation for unaligned Ms
|
|
#pragma unroll
|
|
for (uint32_t s = 0; s < kNumInnerStages; ++ s) {
|
|
// Wait TMA arrivals
|
|
full_barriers[s]->wait((scheduler.current_iter * num_iterations + k_iter) & 1);
|
|
|
|
#pragma unroll
|
|
for (uint32_t local_idx = 0; local_idx < BLOCK_M / WAVE_BLOCK_M; ++ local_idx) {
|
|
auto m_offset = local_idx * WAVE_BLOCK_M;
|
|
auto shifted_accum = accum + WGMMA::kNumAccum * local_idx;
|
|
|
|
// Commit WGMMA instructions
|
|
#pragma unroll
|
|
for (uint32_t i = 0; i < WGMMA::kNumAccum; ++ i)
|
|
warpgroup_fence_operand(accum[i]);
|
|
warpgroup_arrive();
|
|
#pragma unroll
|
|
for (uint32_t k = 0; k < BLOCK_K / WGMMA::K; ++ k) {
|
|
auto desc_a = make_smem_desc(smem_a[s] + (math_wg_idx * WGMMA::M + m_offset) * BLOCK_K + k * WGMMA::K, 1);
|
|
auto desc_b = make_smem_desc(smem_b[s] + k * WGMMA::K, 1);
|
|
WGMMA::wgmma(desc_a, desc_b, shifted_accum, 1);
|
|
}
|
|
warpgroup_commit_batch();
|
|
#pragma unroll
|
|
for (uint32_t i = 0; i < WGMMA::kNumAccum; ++ i)
|
|
warpgroup_fence_operand(accum[i]);
|
|
warpgroup_wait<0>();
|
|
|
|
// Notify barrier arrival at the last warpgroup wave
|
|
if (local_idx == BLOCK_M / WAVE_BLOCK_M - 1)
|
|
empty_barrier_arrive(s);
|
|
}
|
|
}
|
|
|
|
// Wait unaligned cases
|
|
#pragma unroll
|
|
for (uint32_t s = kNumInnerStages; s < kNumStages; ++ s) {
|
|
full_barriers[s]->wait((scheduler.current_iter * num_iterations + k_iter) & 1);
|
|
empty_barrier_arrive(s);
|
|
}
|
|
});
|
|
|
|
// TMA checks
|
|
constexpr uint32_t kNumElemBytes = sizeof(nv_bfloat16);
|
|
constexpr uint32_t TMA_D_BLOCK_N = kSwizzleDMode == 0 ? BLOCK_N : (kSwizzleDMode / kNumElemBytes);
|
|
constexpr uint32_t WGMMA_M_PER_WARP = WGMMA::M / 4;
|
|
DG_STATIC_ASSERT(kSwizzleDMode > 0, "Invalid swizzling type");
|
|
DG_STATIC_ASSERT(BLOCK_M % 8 == 0, "Invalid swizzling atom");
|
|
DG_STATIC_ASSERT(BLOCK_N % TMA_D_BLOCK_N == 0 and BLOCK_N / TMA_D_BLOCK_N <= 32,
|
|
"Unaligned TMA store or too many TMA store instructions");
|
|
DG_STATIC_ASSERT(TMA_D_BLOCK_N % 8 == 0, "Invalid TMA block N");
|
|
|
|
// Wait last TMA store to be finished
|
|
if (threadIdx.x < BLOCK_N / TMA_D_BLOCK_N)
|
|
cute::tma_store_wait<0>();
|
|
cutlass::arch::NamedBarrier(kNumMathThreads).sync();
|
|
|
|
// Write back to shared memory using STSM and issue TMA stores
|
|
DG_STATIC_ASSERT(WGMMA::kNumAccum % 4 == 0, "Invalid STSM x2 vectorization");
|
|
#pragma unroll
|
|
for (uint32_t local_idx = 0; local_idx < BLOCK_M / WAVE_BLOCK_M; ++ local_idx) {
|
|
auto m_offset = local_idx * WAVE_BLOCK_M;
|
|
auto shifted_accum = accum + WGMMA::kNumAccum * local_idx;
|
|
#pragma unroll
|
|
for (auto i = 0; i < WGMMA::kNumAccum / 4; ++ i) {
|
|
// Swizzle or padding into the correct address
|
|
uint8_t* smem_ptr = nullptr;
|
|
if constexpr (kSwizzleDMode > 0) {
|
|
// Calculate the swizzling atom offset and in-atom offset
|
|
constexpr uint32_t kNumBankGroupBytes = 16;
|
|
auto atom_offset = i / (TMA_D_BLOCK_N / 8), in_atom_offset = i % (TMA_D_BLOCK_N / 8);
|
|
|
|
// Calculate the index of the bank group to be written in the atom
|
|
auto bank_group_index = in_atom_offset + lane_idx * (kSwizzleDMode / kNumBankGroupBytes);
|
|
|
|
// Reshape the atom in another view and swizzle
|
|
// - original: `(BLOCK_M, kSwizzleDMode / kNumBankGroupBytes)`
|
|
// - new: `(BLOCK_M * kSwizzleDMode / kNumBankGroupBytes / 8, 8)`
|
|
constexpr bool kHasShortcut = (kSwizzleDMode / kNumBankGroupBytes) == 8;
|
|
auto row = kHasShortcut ? (in_atom_offset / 8 + lane_idx) : (bank_group_index / 8);
|
|
auto col = kHasShortcut ? (in_atom_offset) : (bank_group_index % 8);
|
|
col ^= row % (kSwizzleDMode / 16);
|
|
|
|
// Add back into the base pointer
|
|
// NOTES: think twice before modifying this, as changes may affect the number of instructions
|
|
smem_ptr = reinterpret_cast<uint8_t*>(smem_d) + // Base pointer
|
|
warp_idx * (WGMMA_M_PER_WARP * kSwizzleDMode) + // Warp offset
|
|
m_offset * kSwizzleDMode + // Wave offset
|
|
atom_offset * BLOCK_M * kSwizzleDMode + // Swizzle atom offset (constants)
|
|
row * (kNumBankGroupBytes * 8) + col * kNumBankGroupBytes; // In-atom offset
|
|
} else {
|
|
// No swizzling, just padding
|
|
// TODO: support more cases
|
|
smem_ptr = reinterpret_cast<uint8_t*>(smem_d + (m_offset + warp_idx * WGMMA_M_PER_WARP + lane_idx) * BLOCK_N + i * 8);
|
|
}
|
|
|
|
// NOTES: only 16 lanes' addresses are used
|
|
SM90_U32x2_STSM_N<nv_bfloat162>::copy(
|
|
__float22bfloat162_rn({shifted_accum[i * 4 + 0], shifted_accum[i * 4 + 1]}),
|
|
__float22bfloat162_rn({shifted_accum[i * 4 + 2], shifted_accum[i * 4 + 3]}),
|
|
smem_ptr
|
|
);
|
|
}
|
|
}
|
|
cute::tma_store_fence();
|
|
cutlass::arch::NamedBarrier(kNumMathThreads).sync();
|
|
|
|
// Use TMA store to write back to global memory
|
|
// TODO: compatible with FP32 output
|
|
constexpr bool kWithGroupOffsetD = kGemmType == GemmType::MGroupedMasked;
|
|
DG_STATIC_ASSERT(kNumMathThreads >= BLOCK_N / TMA_D_BLOCK_N, "Too many TMA blocks");
|
|
if (threadIdx.x < BLOCK_N / TMA_D_BLOCK_N) {
|
|
auto in_block_n_offset = threadIdx.x * TMA_D_BLOCK_N;
|
|
auto smem_ptr = smem_d + in_block_n_offset * BLOCK_M;
|
|
cute::SM90_TMA_STORE_2D::copy(&tensor_map_d, smem_ptr,
|
|
n_block_idx * BLOCK_N + in_block_n_offset,
|
|
scheduler.get_global_idx<kWithGroupOffsetD>(shape_m, BLOCK_M, m_block_idx));
|
|
cute::tma_store_arrive();
|
|
}
|
|
__syncwarp();
|
|
}
|
|
}
|
|
#else
|
|
if (blockIdx.x == 0 and threadIdx.x == 0)
|
|
DG_DEVICE_ASSERT(false and "This kernel only support sm_90a");
|
|
#endif
|
|
}
|
|
|
|
}; // namespace deep_gemm
|
|
|
|
#pragma clang diagnostic pop
|